from imgui_bundle import imgui

error       = imgui.ImColor.hsv(0.9667,.88,.64).value
error_bright= imgui.ImColor.hsv(0.9667,.88,.93).value
error_dark  = imgui.ImColor.hsv(0.9667,.88,.43).value

warning     = imgui.ImVec4(1.0000, 193/255, 7/255, 1.)

ok          = imgui.ImVec4(0.0000, 0.8500, 0.0000, 1.)

gray        = imgui.ImVec4(0.5000, 0.5000, 0.5000, 1.)



from shlex import shlex
import shutil
import numpy as np
import pandas as pd
import pathlib
import importlib.resources
import typing

from ... import data_files as _data_files


def _read_glassesValidator_config_file(file):
    # read key=value pairs into dict
    lexer = shlex(file)
    lexer.whitespace += '='
    lexer.wordchars += '.[],'   # don't split extensions of filenames in the input file, and accept stuff from python list syntax
    lexer.commenters = '%'
    return dict(zip(lexer, lexer))

_default_poster_package = '.'.join(__package__.split('.')[:-1]+['default_poster'])

def get_validation_setup(config_dir: str|pathlib.Path=None, config_file: str='validationSetup.txt') -> dict[str,typing.Any]:
    if config_dir is not None:
        with (pathlib.Path(config_dir) / config_file).open() as f:
            validation_config = _read_glassesValidator_config_file(f)
    else:
        # fall back on default config included with package
        with importlib.resources.open_text(_default_poster_package, config_file) as f:
            validation_config = _read_glassesValidator_config_file(f)

    # parse numerics into int or float
    for key,val in validation_config.items():
        if np.all([c.isdigit() for c in val]):
            validation_config[key] = int(val)
        else:
            try:
                validation_config[key] = float(val)
            except:
                pass # just keep value as a string
    return validation_config


def _read_coord_file(config_dir: str|pathlib.Path|None, file: str) -> pd.DataFrame|None:
    if config_dir is not None:
        return _data_files.read_coord_file(pathlib.Path(config_dir) / file)
    else:
        return _data_files.read_coord_file(file, _default_poster_package)

def get_targets(config_dir: str|pathlib.Path=None, file='targetPositions.csv') -> pd.DataFrame|None:
    return _read_coord_file(config_dir, file)

def get_markers(config_dir: str|pathlib.Path=None, file='markerPositions.csv') -> pd.DataFrame|None:
    return _read_coord_file(config_dir, file)



import enum
import pebble
import multiprocessing
import typing
import threading
import dataclasses

from . import utils

ProcessFuture: typing.TypeAlias = pebble.ProcessFuture
_UserDataT = typing.TypeVar("_UserDataT")


class CounterContext:
    _count = -1     # so that first number is 0
    def __enter__(self):
        self._count += 1
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
    def get_count(self):
        return self._count


class State(enum.IntEnum):
    # states for signalling task status, to be stored in files
    Not_Run     = enum.auto()
    Pending     = enum.auto()
    Running     = enum.auto()
    Completed   = enum.auto()
    # two more states needed by process_pool/GUI
    Canceled    = enum.auto()
    Failed      = enum.auto()
    @property
    def displayable_name(self):
        return self.name.replace("_", " ")
utils.register_type(utils.CustomTypeEntry(State,'__enum.process.State__', utils.enum_val_2_str, lambda x: getattr(State, x.split('.')[1])))


class ProcessWaiter(object):
    """Routes completion through to user callback."""
    def __init__(self, job_id: int, user_data: _UserDataT, done_callback: typing.Callable[[ProcessFuture, _UserDataT, int, State], None]):
        self.done_callback  = done_callback
        self.job_id         = job_id
        self.user_data      = user_data

    def add_result(self, future: ProcessFuture):
        self._notify(future, State.Completed)

    def add_exception(self, future: ProcessFuture):
        self._notify(future, State.Failed)

    def add_cancelled(self, future: ProcessFuture):
        self._notify(future, State.Canceled)

    def _notify(self, future: ProcessFuture, state: State):
        self.done_callback(future, self.job_id, self.user_data, state)

class PoolJob(typing.NamedTuple):
    future   : ProcessFuture
    user_data: _UserDataT
class ProcessPool:
    def __init__(self, num_workers = 2):
        self.num_workers            = num_workers
        self.auto_cleanup_if_no_work= False

        # NB: pool is only started in run() once needed
        self._pool              : pebble.pool.ProcessPool   = None
        self._jobs              : dict[int,PoolJob]         = None
        self._job_id_provider   : CounterContext            = CounterContext()
        self._lock              : threading.Lock            = threading.Lock()

    def _cleanup(self):
        # cancel all pending and running jobs
        self.cancel_all_jobs()

        # stop pool
        if self._pool and self._pool.active:
            self._pool.stop()
            self._pool.join()
        self._pool = None
        self._jobs = None

    def cleanup(self):
        with self._lock:
            self._cleanup()

    def cleanup_if_no_jobs(self):
        with self._lock:
            self._cleanup_if_no_jobs()

    def _cleanup_if_no_jobs(self):
        # NB: lock must be acquired when calling this
        if self._pool and not self._jobs:
            self._cleanup()

    def set_num_workers(self, num_workers: int):
        # NB: doesn't change number of workers on an active pool, only takes effect when pool is restarted
        self.num_workers = num_workers

    def run(self, fn: typing.Callable, user_data: _UserDataT=None, done_callback: typing.Callable[[ProcessFuture, _UserDataT, int, State], None]=None, *args, **kwargs) -> tuple[int, ProcessFuture]:
        with self._lock:
            if self._pool is None or not self._pool.active:
                context = multiprocessing.get_context("spawn")  # ensure consistent behavior on Windows (where this is default) and Unix (where fork is default, but that may bring complications)
                self._pool = pebble.ProcessPool(max_workers=self.num_workers, context=context)

            if self._jobs is None:
                self._jobs = {}

            with self._job_id_provider:
                job_id = self._job_id_provider.get_count()
                self._jobs[job_id] = PoolJob(self._pool.schedule(fn, args=args, kwargs=kwargs), user_data)
                self._jobs[job_id].future._waiters.append(ProcessWaiter(job_id, user_data, self._job_done_callback))
                if done_callback:
                    self._jobs[job_id].future._waiters.append(ProcessWaiter(job_id, user_data, done_callback))
                return job_id, self._jobs[job_id].future

    def _job_done_callback(self, future: ProcessFuture, job_id: int, user_data: _UserDataT, state: State):
        with self._lock:
            if self._jobs is not None and job_id in self._jobs:
                # clean up the work item since we're done with it
                del self._jobs[job_id]

            if self.auto_cleanup_if_no_work:
                # close pool if no work left
                self._cleanup_if_no_jobs()

    def get_job_state(self, job_id: int) -> State:
        if not self._jobs:
            return None
        job = self._jobs.get(job_id, None)
        if job is None:
            return None
        else:
            return _get_status_from_future(job.future)

    def get_job_user_data(self, job_id: int) -> _UserDataT:
        if not self._jobs:
            return None
        job = self._jobs.get(job_id, None)
        if job is None:
            return None
        else:
            return job.user_data

    def cancel_job(self, job_id: int) -> bool:
        if not self._jobs:
            return False
        if (job := self._jobs.get(job_id, None)) is None:
            return False
        return job.future.cancel()

    def cancel_all_jobs(self):
        if not self._jobs:
            return
        for job_id in reversed(self._jobs): # reversed so that later pending jobs don't start executing when earlier gets cancelled, only to be canceled directly after
            if not self._jobs[job_id].future.done():
                self._jobs[job_id].future.cancel()


class JobPayload(typing.NamedTuple):
    fn:     typing.Callable[..., None]
    args:   tuple
    kwargs: dict

@dataclasses.dataclass
class JobDescription(typing.Generic[_UserDataT]):
    user_data:          _UserDataT
    payload:            JobPayload
    done_callback:      typing.Callable[[ProcessFuture, _UserDataT, int, State], None]

    exclusive_id:       typing.Optional[int]      = None # if set, only one task with a given id can be run at a time, rest are kept in waiting. E.g. to ensure only one task needing a gui is run at a time
    priority:           int                       = 999  # jobs with a higher priority are scheduled first, unless they cannot be because they're exclusive (exclusive_id is set) and task of that exclusivity is already running, or because their dependencies are not met yet
    depends_on:         typing.Optional[set[int]] = None # set of job ids that need to be completed before this one can be launched

    _pool_job_id:       typing.Optional[int]            = None
    _future:            typing.Optional[ProcessFuture]  = dataclasses.field(init=False, default=None)
    _final_state:       typing.Optional[State]          = dataclasses.field(init=False, default=None)
    error:              typing.Optional[str]            = None

    def get_state(self) -> State:
        if self._final_state is not None:
            return self._final_state
        elif self._future is not None:
            job_state = _get_status_from_future(self._future)
            if job_state not in [State.Pending, State.Running]:
                # finished, cache result
                self._final_state = job_state
                # can also dump the future
                self._future = None
            return job_state
        return State.Pending

    def is_scheduled(self):
        # True if job is scheduled to the pool
        return self._pool_job_id is not None and self.get_state() in [State.Pending, State.Running]

class JobScheduler(typing.Generic[_UserDataT]):
    def __init__(self, pool: ProcessPool, job_is_valid_checker : typing.Callable[[_UserDataT], bool]|None = None):
        self.jobs               : dict[int, JobDescription[_UserDataT]] = {}
        self._job_id_provider   : CounterContext                        = CounterContext()
        self._pending_jobs      : list[int]                             = []    # jobs not scheduled or finished

        self._job_is_valid_checker  = job_is_valid_checker
        self._pool                  = pool

    def add_job(self,
                user_data: _UserDataT, payload: JobPayload, done_callback: typing.Callable[[ProcessFuture, _UserDataT, int, State], None],
                exclusive_id: typing.Optional[int] = None, priority: int = None, depends_on: typing.Optional[set[int]] = None) -> int:
        with self._job_id_provider:
            job_id = self._job_id_provider.get_count()
        self.jobs[job_id] = JobDescription(user_data, payload, done_callback, exclusive_id, priority, depends_on)
        self._pending_jobs.append(job_id)
        return job_id

    def cancel_job(self, job_id: int):
        if job_id not in self.jobs:
            return

        if self.jobs[job_id]._pool_job_id is not None:
            self._pool.cancel_job(self.jobs[job_id]._pool_job_id)
        else:
            if job_id in self._pending_jobs:
                self._pending_jobs.remove(job_id)
            self.jobs[job_id]._final_state = State.Canceled
            if self.jobs[job_id].done_callback:
                self.jobs[job_id].done_callback(None, None, self.jobs[job_id].user_data, State.Canceled)
        # TODO: also cancel all jobs that depend on this job

    def cancel_all_jobs(self):
        # cancel any jobs that may still be running
        for job_id in self.jobs:
            if self.jobs[job_id]._final_state not in [State.Completed, State.Canceled, State.Failed]:
                self.cancel_job(job_id)
        # make double sure they're cancelled
        self._pool.cancel_all_jobs()
        # ensure pool is no longer running
        self._pool.cleanup()

    def clear(self):
        # cancel any jobs that may still be running
        self.cancel_all_jobs()
        # clean up
        self._pending_jobs.clear()
        self.jobs.clear()
        # reset job counter
        self._job_id_provider = CounterContext()

    def update(self):
        # first count how many are scheduled to the pool and whether all tasks are still valid
        num_scheduled_to_pool = 0
        exclusive_ids: set[int] = set()
        for job_id in self.jobs:
            job = self.jobs[job_id]
            # check job still valid or should be canceled
            if self._job_is_valid_checker is not None:
                if not self._job_is_valid_checker(job.user_data):
                    self.cancel_job(job_id)
            # remove finished job from pending jobs if its still there
            if job._final_state is not None and job_id in self._pending_jobs:
                self._pending_jobs.remove(job_id)
            # check how many scheduled jobs we have
            if job.is_scheduled():
                num_scheduled_to_pool += 1
                if job.exclusive_id is not None:
                    exclusive_ids.add(job.exclusive_id)

        # if we have less than max number of tasks scheduled to the pool, see if anything new to schedule to the pool
        while num_scheduled_to_pool < self._pool.num_workers:
            # find suitable next task to schedule
            # order tasks by priority, filtering out those who have a colliding exclusive_id
            job_ids = [i for i in sorted(self._pending_jobs, key=lambda ii: 999 if self.jobs[ii].priority is None else self.jobs[ii].priority) if self.jobs[i].exclusive_id not in exclusive_ids]
            if not job_ids:
                break
            job_id = job_ids[0]
            to_schedule = self.jobs[job_id]

            to_schedule._pool_job_id, to_schedule._future = \
                self._pool.run(to_schedule.payload.fn, to_schedule.user_data, to_schedule.done_callback, *to_schedule.payload.args, **to_schedule.payload.kwargs)
            self._pending_jobs.remove(job_id)
            if to_schedule.exclusive_id is not None:
                exclusive_ids.add(to_schedule.exclusive_id)
            num_scheduled_to_pool += 1


def _get_status_from_future(fut: ProcessFuture) -> State:
    if fut.running():
        return State.Running
    elif fut.done():
        if fut.cancelled():
            return State.Canceled
        elif fut.exception() is not None:
            return State.Failed
        else:
            return State.Completed
    else:
        return State.Pending


import enum

import numpy as np
from matplotlib import colors
import pathlib
import typing
import math
import cv2

from .. import drawing as _drawing, marker as _marker, plane as _plane, transforms as _transforms, utils as _utils

from . import config
from . import default_poster

# NB: using pose information requires a calibrated scene camera
class DataQualityType(enum.Enum):
    viewpos_vidpos_homography   = enum.auto()   # use homography to map gaze from video to poster, and viewing distance defined in config (combined with the assumptions that the viewing position (eye) is located directly in front of the poster's center and that the poster is oriented perpendicularly to the line of sight) to compute angular measures
    pose_vidpos_homography      = enum.auto()   # use homography to map gaze from video to poster, and pose information w.r.t. poster to compute angular measures
    pose_vidpos_ray             = enum.auto()   # use camera calibration to map gaze position on scene video to cyclopean gaze vector, and pose information w.r.t. poster to compute angular measures
    pose_world_eye              = enum.auto()   # use provided gaze position in world (often a binocular gaze point), and pose information w.r.t. poster to compute angular measures
    pose_left_eye               = enum.auto()   # use provided left eye gaze vector, and pose information w.r.t. poster to compute angular measures
    pose_right_eye              = enum.auto()   # use provided right eye gaze vector, and pose information w.r.t. poster to compute angular measures
    pose_left_right_avg         = enum.auto()   # report average of left (pose_left_eye) and right (pose_right_eye) eye angular values

    # so this gets serialized in a user-friendly way by pandas..
    def __str__(self):
        return self.name
_utils.register_type(_utils.CustomTypeEntry(DataQualityType, 'glassesValidator.DataQualityType', _utils.enum_val_2_str, lambda x: _utils.enum_str_2_val(x, DataQualityType)))

def get_DataQualityType_explanation(dq: DataQualityType):
    ler_name =  "Left eye ray + pose"
    rer_name = "Right eye ray + pose"
    match dq:
        case DataQualityType.viewpos_vidpos_homography:
            return "Homography + view distance", \
                   "Use homography to map gaze position from the scene video to " \
                   "the validation poster, and use an assumed viewing distance (see " \
                   "the project's configuration) to compute data quality measures " \
                   "in degrees with respect to the scene camera. In this mode, it is "\
                   "assumed that the eye is located exactly in front of the center of "\
                   "the poster and that the poster is oriented perpendicularly to the "\
                   "line of sight from this assumed viewing position."
        case DataQualityType.pose_vidpos_homography:
            return "Homography + pose", \
                   "Use homography to map gaze position from the scene video to " \
                   "the validation poster, and use the determined pose of the scene " \
                   "camera (requires a calibrated camera) to compute data quality " \
                   "measures in degrees with respect to the scene camera."
        case DataQualityType.pose_vidpos_ray:
            return "Video ray + pose", \
                   "Use camera calibration to turn gaze position from the scene " \
                   "video into a direction vector, and determine gaze position on " \
                   "the validation poster by intersecting this vector with the " \
                   "poster. Then, use the determined pose of the scene camera " \
                   "(requires a calibrated camera) to compute data quality " \
                   "measures in degrees with respect to the scene camera."
        case DataQualityType.pose_world_eye:
            return "World gaze position + pose", \
                   "Use the gaze position in the world provided by the eye tracker " \
                   "(often a binocular gaze point) to determine gaze position on the " \
                   "validation poster by turning it into a direction vector with respect " \
                   "to the scene camera and intersecting this vector with the poster. " \
                   "Then, use the determined pose of the scene camera " \
                   "(requires a calibrated camera) to compute data quality " \
                   "measures in degrees with respect to the scene camera."
        case DataQualityType.pose_left_eye:
            return ler_name, \
                   "Use the gaze direction vector for the left eye provided by " \
                   "the eye tracker to determine gaze position on the validation " \
                   "poster by intersecting this vector with the poster. " \
                   "Then, use the determined pose of the scene camera " \
                   "(requires a camera calibration) to compute data quality " \
                   "measures in degrees with respect to the left eye."
        case DataQualityType.pose_right_eye:
            return rer_name, \
                   "Use the gaze direction vector for the right eye provided by " \
                   "the eye tracker to determine gaze position on the validation " \
                   "poster by intersecting this vector with the poster. " \
                   "Then, use the determined pose of the scene camera " \
                   "(requires a camera calibration) to compute data quality " \
                   "measures in degrees with respect to the right eye."
        case DataQualityType.pose_left_right_avg:
            return "Average eye rays + pose", \
                   "For each time point, take angular offset between the left and " \
                   "right gaze positions and the fixation target and average them " \
                   "to compute data quality measures in degrees. Requires " \
                   f"'{ler_name}' and '{rer_name}' to be enabled."


class Plane(_plane.Plane):
    poster_image_filename = 'referencePoster.png'
    default_aruco_dict    = cv2.aruco.DICT_4X4_250

    def __init__(self, config_dir: str|pathlib.Path|None, validation_config: dict[str,typing.Any]=None, **kwarg):
        # NB: if config_dir is None, the default config will be used

        if config_dir is not None:
            config_dir = pathlib.Path(config_dir)

        # get validation config, if needed
        if validation_config is None:
            validation_config = config.get_validation_setup(config_dir)
        self.config = validation_config

        # get marker width
        if self.config['mode'] == 'deg':
            self.cell_size_mm = 2.*math.tan(math.radians(.5))*self.config['distance']*10
        else:
            self.cell_size_mm = 10 # 1cm
        markerSize = self.cell_size_mm*self.config['markerSide']

        # get board size
        plane_size = _plane.Coordinate(self.config['gridCols']*self.cell_size_mm, self.config['gridRows']*self.cell_size_mm)

        # get targets first, so that they can be drawn on the reference image
        self.targets: dict[int,_marker.Marker] = {}
        origin = self._get_targets(config_dir, self.config)

        # call base class
        markers = config.get_markers(config_dir, self.config['markerPosFile'])
        ref_image_store_path = None
        if 'ref_image_store_path' in kwarg:
            ref_image_store_path = kwarg.pop('ref_image_store_path')
        elif config_dir is not None:
            ref_image_store_path = config_dir / self.poster_image_filename
        super(Plane, self).__init__(markers, markerSize, plane_size, Plane.default_aruco_dict, self.config['markerBorderBits'],self.cell_size_mm, "mm", ref_image_store_path=ref_image_store_path, ref_image_size=self.config['referencePosterSize'],**kwarg)

        # set center
        self.set_origin(origin)

    def set_origin(self, origin: _plane.Coordinate):
        # set origin of plane. Origin location is on current (not original) plane
        # so set_origin([5., 0.]) three times in a row shifts the origin rightward by 15 units
        for i in self.targets:
            self.targets[i].shift(-np.array(origin))
        super(Plane, self).set_origin(origin)

    def _get_targets(self, config_dir, validationSetup) -> _plane.Coordinate:
        """ poster space: (0,0) is origin (might be center target), (-,-) bottom left """

        # read in target positions
        targets = config.get_targets(config_dir, validationSetup['targetPosFile'])
        if targets is not None:
            targets['center'] = list(targets[['x','y']].values)
            targets['center'] *= self.cell_size_mm
            targets = targets.drop(['x','y'], axis=1)
            self.targets = {idx:_marker.Marker(idx,**kwargs) for idx,kwargs in zip(targets.index.values,targets.to_dict(orient='records'))}
            origin = _plane.Coordinate(*targets.loc[validationSetup['centerTarget']].center.copy())  # NB: need origin in scaled space
        else:
            origin = _plane.Coordinate(0.,0.)
        return origin

    def _store_reference_image(self, path: pathlib.Path, width: int) -> np.ndarray:
        # first call superclass method to generate image without targets
        img = super(Plane, self)._store_reference_image(path, width)
        height = img.shape[0]

        # add targets
        subPixelFac = 8   # for sub-pixel positioning
        for key in self.targets:
            # 1. determine position on image
            circlePos = _transforms.to_image_pos(*self.targets[key].center, self.bbox,[width,height])

            # 2. draw
            clr = tuple([int(i*255) for i in (colors.to_rgb(self.targets[key].color)[::-1] if self.targets[key].color else (0.,0.,1.))])  # need BGR color ordering
            _drawing.openCVCircle(img, circlePos, 15, clr, -1, subPixelFac)

        if path:
            cv2.imwrite(path, img)

        return img